from . import log
logger = log.get(__name__)

class Processor:
    
    def __init__(self, registry):
        self._registry = registry
        self._messageHandlers = {
            "policies": self._acceptPolicies,
            "endpoints": self._acceptEndpoints,
            "peers": self._acceptPeers
        }

    def accept(self, section, data):
        logger.debug('Accept Section: %s, data: %s', section, data)

        handler = self._messageHandlers[section]
        if handler is not None: 
            return handler(data)

    def _acceptPolicies(self, data):
        if data is not None:
            self._registry.set('policies', [], data)
        else:
            self._registry.reset('policies')

    def _acceptEndpoints(self, data):
        if data is not None:
            self._registry.set('endpoints', [], data)
            for endpointName, endpoints in data.items():
                self._registry.set('endpoints', [endpointName], endpoints)
        else:
            self._registry.reset('endpoints')

    def _acceptPeers(self, data):
        if data is None:
            self._registry.reset('peer')
        else:
            self._registry.set('endpoints', [], data)
            for serviceId, serviceData in data.items():
                if self._isEndpointService(serviceId):
                    self._handleServicePeers(serviceId, serviceData)
                else:
                    self._handleResourcePeers(serviceId, serviceData)
            
    def _isEndpointService(self, serviceId):
        return serviceId.startswith('service://') or serviceId.startswith('cluster://')

    def _handleServicePeers(self, serviceId, serviceData):
        for endpoint, endpointData in serviceData.items():
            self._registry.set('peer', [serviceId, endpoint], endpointData)

    def _handleResourcePeers(self, resourceId, resourceData):
        self._registry.set('peer', [resourceId], resourceData)



from . import log
logger = log.get(__name__)

logger.info('Starting...')

import sys
import json
import random as rand

from .registry import Registry
from .policy import Policy
from .zipkin import Zipkin
from .processor import Processor
from .client import Client
from .executor import Executor
from .deep_eq import deep_eq

import copy
import os

berlioz_cluster = os.environ.get('BERLIOZ_CLUSTER')
berlioz_sector = os.environ.get('BERLIOZ_SECTOR')
berlioz_service = os.environ.get('BERLIOZ_SERVICE')


registry = Registry()
policy = Policy(registry)
zipkin = Zipkin(policy)
processor = Processor(registry)

def onMessage(msg):
    for section, data in msg.items():
        # logger.info('Section %s, data: %s', section, data)
        processor.accept(section, data)
    # logger.info('**** REGISTRY: %s', json.dumps(registry.extractRoot(), indent=4, sort_keys=True))
client = Client(onMessage)


def monitorPeers(peerPath, cb):
    registry.subscribe('peer', peerPath, cb)


def monitorPeer(peerPath, selector, cb):
    class nonlocal:
        oldValue = None
    
    def innerCb(peers):
        value = selector(peers)
        isChanged = False
        if value is not None:
            if nonlocal.oldValue is not None:
                isChanged = not deep_eq(value, nonlocal.oldValue)
            else:
                isChanged = True
        else:
            if nonlocal.oldValue is not None:
                isChanged = True
            else:
                isChanged = False
        if isChanged:
            nonlocal.oldValue = value
            cb(value)

    registry.subscribe('peer', peerPath, innerCb)

def getPeers(peerPath):
    return registry.get('peer', peerPath)

def getPeer(peerPath, selector):
    peers = registry.get('peer', peerPath)
    return selector(peers)

def selectFirstPeer(peers):
    return firstFromDict(peers)

def selectRandomPeer(peers):
    return randomFromDict(peers)





def monitorNatives(kind, name, cb):
    registry.subscribe(kind, [name], cb)

def getNatives(kind, name):
    return registry.get(kind, [name])

def getNative(kind, name):
    peers = getNatives(kind, name)
    return randomFromDict(peers)





def instrument(method, binary_annotations):
    return zipkin.instrument(method, binary_annotations)


def randomFromList(list):
    return rand.choice(list)

def firstFromList(list):
    return list[0]

def randomFromDict(dict):
    if not dict:
        return None
    key = randomFromList(list(dict.keys()))
    return dict[key]

def firstFromDict(dict):
    if not dict:
        return None
    key = firstFromList(list(dict.keys()))
    return dict[key]

def setupFlask(app):
    from .frameworks.b_flask import Flask
    Flask(app, zipkin, policy)


from . import log
logger = log.get(__name__)

from .service import Service

class Sector:
    def __init__(self, starter, name):
        self._starter = starter
        self._name = name

    def service(self, name, endpoint):
        id = 'service://' + '-'.join([self._starter.berlioz_cluster, self._name, name])
        return Service(self._starter, id, endpoint)


from flask import g as flask_g
from flask import request as flask_request

class Flask:
    def __init__(self, app, zipkin, policy):
        self._app = app
        self._zipkin = zipkin
        self._policy = policy
        
        self._app.before_request(self._handleBeforeRequest)
        self._app.after_request(self._handleAfterRequest)

    def _handleBeforeRequest(self):
        url = ''
        if flask_request.script_root is not None:
           url = flask_request.script_root
        if flask_request.path is not None:
           url = url + flask_request.path
        zipkin_span = self._zipkin.instrumentServer(flask_request.headers,
                                                    flask_request.method,
                                                    url)
        flask_g._berlioz_zipkin_span = zipkin_span
        zipkin_span.start()

    def _handleAfterRequest(self, response):
        zipkin_span = getattr(flask_g, '_berlioz_zipkin_span')
        if zipkin_span:
            self._zipkin.serverResponse(zipkin_span, response.status_code)
        return response


from . import log
logger = log.get(__name__)

from .peer_accessor import PeerAccessor


from . import aws as AWS
nativeClientFetcher = {
    "dynamodb" : AWS.fetchDynamoClient,
    "kinesis" : AWS.fetchKinesisClient,
    "rsa-secret" : AWS.fetchSSMClient
}
nativeClientArgSetter = {
    "kinesis" : AWS.setupKinesisArgs,
    "rsa-secret" : AWS.setupParameterArgs
}

class NativeResource(PeerAccessor):
    def __init__(self, starter, id, endpoint):
        if endpoint is None:
            endpoint = 'default'
        PeerAccessor.__init__(self, starter, [id, endpoint])

    def client(self):
        nativeClient = NativeResourceWrapper(self)
        return nativeClient

class NativeResourceWrapper(object):

    def __init__(self, target):
        object.__setattr__(self, "_target", target)

    def __getattribute__(self, propKey):

        def perform(*args, **kwargs):
            target = object.__getattribute__(self, "_target")

            def execAction(peer, zipkin_span=None):
                logger.info('Running %s', propKey)
                clientFetcher = nativeClientFetcher.get(peer['subClass'])
                if not clientFetcher:
                    raise Exception('Service not supported', peer['subClass'])
                client = clientFetcher(peer)
                origMethod = getattr(client, propKey)
                argsSetter = nativeClientArgSetter.get(peer['subClass'])
                if argsSetter:
                    argsSetter(peer, propKey, args, kwargs)
                logger.info('Args: %s ', args)
                logger.info('kwargs: %s ', kwargs)
                result = origMethod(*args, **kwargs)
                logger.info('Running %s completed', propKey)
                return result

            return target.performExecutor(propKey, None, execAction)

        return perform



from . import log
logger = log.get(__name__)

from .executor import Executor

class PeerAccessor:
    def __init__(self, starter, peerPath):
        self._starter = starter
        self._peerPath = peerPath

    @property
    def peerPath(self):
        return self._peerPath

    @property
    def peerId(self):
        return self._peerPath[0]

    @property
    def registry(self):
        return self._starter.registry
    
    def monitorAll(self, cb):
        return self._starter.monitorPeers(self.peerPath, cb)

    def monitorFirst(self, cb):
        return self._starter.monitorPeer(self.peerPath, self._starter.selectFirstPeer, cb)

    def all(self):
        return self._starter.getPeers(self.peerPath)

    def first(self):
        return self._starter.getPeer(self.peerPath, self._starter.selectFirstPeer)

    def random(self):
        return self._starter.getPeer(self.peerPath, self._starter.selectRandomPeer)

    def performExecutor(self, trackerMethod, binary_annotations, actionCb):
        executor = Executor(self, self._starter.policy, self._starter.zipkin, self.peerPath, trackerMethod, binary_annotations, actionCb)
        return executor.perform()


import os
from setuptools import setup

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

from berlioz import metadata

setup(
    name = "berlioz",
    version = metadata.VERSION,
    url = 'https://berlioz.cloud',
    author = "Berlioz",
    author_email = "info@berlioz.cloud",
    maintainer = "Ruben Hakopian",
    maintainer_email = "ruben.hakopian@gmail.com",
    description = ("Python SDK for Berlioz Cloud"),
    license = "Apache License 2.0",
    keywords = "microservices service mesh cloud aws",
    packages=['berlioz', 'berlioz/frameworks'],
    include_package_data=True,
    long_description=read('README.rst'),
    python_requires='>=2.7,>=3.6',
    install_requires=[
        'boto3>=1.7.41',
        'requests>=2.19.1',
        'websocket-client>=0.48.0',
        'py_zipkin>=0.12.0'
    ],
    platforms='any',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Flask",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Monitoring"
    ],
)



import logging

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(levelname)s - %(name)s - %(message)s')

# console = logging.StreamHandler()
# console.setLevel(logging.DEBUG)
# formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(name)s - %(message)s')
# console.setFormatter(formatter)
# logging.getLogger('').addHandler(console)

def get(name):
    logger = logging.getLogger(name)
    logger.info("Logger Init")
    return logger